import logging
from typing import List, Callable

from octane_sdk_wrapper import Octane, OctaneReaderMode, OctaneSearchMode, OctaneTagReport

from . import SenseidReader, SenseidReaderDetails
from ..parsers import SenseidTag
from ..parsers.rain import SenseidRainTag

logger = logging.getLogger(__name__)


class SenseidOctane(SenseidReader):

    def __init__(self):
        self.driver = Octane()
        self.notification_callback = None
        self.details = None

    def connect(self, connection_string: str):
        if not self.driver.connect(ip=connection_string):
            return False
        self.driver.set_notification_callback(self._octane_notification_callback)
        self.get_details()
        # Set Senseid compatible mode
        self.driver.set_mode(reader_mode=OctaneReaderMode.DenseReaderM4, search_mode=OctaneSearchMode.DualTarget,
                             session=1)
        # Set MAX TX Power
        self.driver.set_tx_power(self.details.max_tx_power)
        # Enable first antenna
        antenna_config = [False] * self.details.antenna_count
        antenna_config[0] = True
        self.driver.set_antenna_config(antenna_config)
        return True

    def _octane_notification_callback(self, octane_tag_report: OctaneTagReport):
        if self.notification_callback is not None:
            self.notification_callback(SenseidRainTag(epc=octane_tag_report.Epc))

    def disconnect(self):
        self.driver.disconnect()

    def get_details(self) -> SenseidReaderDetails:
        if self.details is None:
            feature_set = self.driver.query_feature_set()
            self.details = SenseidReaderDetails(
                model_name=feature_set.model_name,
                region=feature_set.region,
                firmware_version=feature_set.firmware_version,
                antenna_count=feature_set.antenna_count,
                min_tx_power=feature_set.min_tx_power,
                max_tx_power=feature_set.max_tx_power
            )
        return self.details

    def get_tx_power(self) -> float:
        # Only supporting same power on all antennas
        return self.driver.get_tx_power()[0]

    def set_tx_power(self, dbm: float):
        # Only supporting same power on all antennas
        if self.details is None:
            self.get_details()
        if dbm > self.details.max_tx_power:
            dbm = self.details.max_tx_power
            logger.warning('Power set to max power: ' + str(dbm))
        if dbm < self.details.min_tx_power:
            dbm = self.details.min_tx_power
            logger.warning('Power set to min power: ' + str(dbm))
        self.driver.set_tx_power(dbm=dbm)

    def get_antenna_config(self) -> List[bool]:
        antenna_config_array = self.driver.get_antenna_config()
        return antenna_config_array

    def set_antenna_config(self, antenna_config_array: List[bool]):
        if not (True in antenna_config_array):
            antenna_config_array[0] = True
            logger.warning('At least one antenna needs to be active. Enabling antenna 1.')
        self.driver.set_antenna_config(antenna_config_array)

    def start_inventory_async(self, notification_callback: Callable[[SenseidTag], None]):
        self.notification_callback = notification_callback
        return self.driver.start()

    def stop_inventory_async(self):
        return self.driver.stop()



import logging
from dataclasses import dataclass
from enum import Enum

from dataclasses_json import dataclass_json

from .. import SenseidData, SenseidTag, SenseidTechnologies

logger = logging.getLogger(__name__)


class Endianness(Enum):
    LITTLE = False
    BIG = True


URI_PREFIXES = {
    0x00: "",
    0x01: "http://www.",
    0x02: "https://www.",
    0x03: "http://",
    0x04: "https://",
    0x05: "tel:",
    0x06: "mailto:",
    0x07: "ftp://anonymous:anonymous@",
    0x08: "ftp://ftp.",
    0x09: "ftps://",
    0x0A: "sftp://",
    0x0B: "smb://",
    0x0C: "nfs://",
    0x0D: "ftp://",
    0x0E: "dav://",
    0x0F: "news:",
    0x10: "telnet://",
    0x11: "imap:",
    0x12: "rtsp://",
    0x13: "urn:",
    0x14: "pop:",
    0x15: "sip:",
    0x16: "sips:",
    0x17: "tftp:",
    0x18: "btspp://",
    0x19: "btl2cap://",
    0x1A: "btgoep://",
    0x1B: "tcpobex://",
    0x1C: "irdaobex://",
    0x1D: "file://",
    0x1E: "urn:epc:id:",
    0x1F: "urn:epc:tag:",
    0x20: "urn:epc:pat:",
    0x21: "urn:epc:raw:",
    0x22: "urn:epc:",
    0x23: "urn:nfc:",
}


def convert_to_uint(data, uint_size, endianness):
    """
    Converts raw bytes into integers according to size and endianness.
    - data: bytearray or list of bytes
    - uint_size: number of bytes per integer (e.g. 2 for uint16, 4 for uint32)
    - endianness: True for big-endian, False for little-endian
    """
    if data is None:
        return None

    result = []
    for i in range(0, len(data), uint_size):
        byte_group = list(data[i:i + uint_size])
        if len(byte_group) < uint_size:
            break
        if endianness == Endianness.LITTLE.value:
            byte_group.reverse()
        value = 0
        for byte_val in byte_group:
            value = (value << 8) | byte_val
        result.append(value)

    return result


@dataclass_json
@dataclass
class SenseidNfcTag(SenseidTag):

    def __init__(self, ndef_data: bytearray, uid: str = None):
        self.technology = SenseidTechnologies.NFC
        self.parse_ndef(ndef_data, uid)

    def parse_ndef(self, ndef_data: bytearray, uid: str = None):
        if ndef_data is None or len(ndef_data) < 11:
            self._set_unknown(uid)
            logger.debug(f"NDEF packet too short: {len(ndef_data) if ndef_data else 0} bytes")
            return

        # Verify CC file (E1 40)
        if ndef_data[0] != 0xE1 or ndef_data[1] != 0x40:
            self._set_unknown(uid)
            logger.debug(f"Invalid CC file: {ndef_data[0]:02X} {ndef_data[1]:02X}")
            return

        # Verify NDEF TLV
        if ndef_data[4] != 0x03:
            self._set_unknown(uid)
            logger.debug(f"Invalid NDEF TLV: {ndef_data[4]:02X}")
            return

        # Verify type 'U' (URI)
        if ndef_data[9] != 0x55:
            self._set_unknown(uid)
            logger.debug(f"NDEF type is not URI: {ndef_data[9]:02X}")
            return

        # URI Identifier Code
        uri_prefix = URI_PREFIXES.get(ndef_data[10], "")

        # Extract payload (URL)
        payload_length = ndef_data[8]
        payload_start = 11
        payload_end = payload_start + payload_length - 1  # -1 because URI ID is already counted

        if payload_end > len(ndef_data):
            payload_end = len(ndef_data)

        url_part = ''.join(chr(b) for b in ndef_data[payload_start:payload_end] if b != 0xFE and b != 0x00)
        full_url = uri_prefix + url_part

        # Set tag identity
        self.id = uid or ''
        self.fw_version = None
        self.sn = None
        self.name = 'NFC NDEF Tag'
        self.description = full_url

        # Extract temperature and humidity from URL format: IP:PORT/TEMP,HUM
        self.data = self._parse_sensor_data(url_part)
        logger.debug('Parsing done -> ' + str(self))

    def _parse_sensor_data(self, url_part):
        try:
            if '/' in url_part:
                data_part = url_part.split('/')[-1]
                if ',' in data_part:
                    temp_str, hum_str = data_part.split(',')
                    temperature = int(temp_str) / 100.0
                    humidity = int(hum_str) / 100.0
                    return [
                        SenseidData(
                            magnitude='Temperature',
                            unit_long='Celsius',
                            unit_short='°C',
                            value=temperature
                        ),
                        SenseidData(
                            magnitude='Humidity',
                            unit_long='Percent',
                            unit_short='%',
                            value=humidity
                        ),
                    ]
        except (ValueError, IndexError) as e:
            logger.debug(f"Error extracting sensor data: {e}")
        return None

    def _set_unknown(self, uid):
        self.id = uid or ''
        self.fw_version = None
        self.sn = None
        self.name = 'NFC Tag'
        self.description = 'Unknown NFC tag'
        self.data = None



import logging
import struct
from dataclasses import dataclass
from math import log

from dataclasses_json import dataclass_json

from .yaml import SENSEID_BLE_DEF, SenseidValueType, SenseidTransformType
from .. import SenseidData, SenseidTag, SenseidTechnologies

logger = logging.getLogger(__name__)


@dataclass_json
@dataclass
class SenseidBleTag(SenseidTag):

    def __init__(self, beacon: str | bytearray):
        self.technology = SenseidTechnologies.BLE
        self.parse_beacon(beacon)

    def _get_bytearray_beacon(self, beacon: str | bytearray):
        if not (isinstance(beacon, str) or isinstance(beacon, bytearray)):
            raise TypeError('epc must be a hex string or bytearray')
        if isinstance(beacon, str):
            try:
                return bytearray.fromhex(beacon)
            except Exception as e:
                raise TypeError('epc must be a hex string or bytearray')
        return beacon

    def _is_senseid_beacon(self, beacon_bytes: bytearray):
        try:
            local_name = beacon_bytes[6+2:len(SENSEID_BLE_DEF.local_name)+6+2]
            if local_name.decode() != SENSEID_BLE_DEF.local_name:
                return False
            if len(beacon_bytes) < len(SENSEID_BLE_DEF.local_name) + 2 + 3:  # PEN + TYPE + SN
                return False
            return True
        except Exception as e:
            return False

    def _parse_senseid_beacon(self, beacon_bytes: bytearray):
        senseid_type = beacon_bytes[6+8]
        fw_version = beacon_bytes[6+9]
        #senseid_sn_bytes = beacon_bytes[7:10]
        id = beacon_bytes[0:6].hex(sep=':').upper()

        #sn = struct.unpack('>I', bytearray([0]) + senseid_sn_bytes)[0]
        sn = 0
        if senseid_type not in SENSEID_BLE_DEF.types:
            self.id = 'ID'
            self.fw_version = None
            self.sn = None
            self.name = 'Unknown SenseID type'
            self.description = 'Unknown SenseID type'
            self.data = None
            return

        senseid_value_bytes = beacon_bytes[6+10:]
        senseid_type_config = SENSEID_BLE_DEF.types[senseid_type]
        self.id = id
        self.fw_version = fw_version
        self.sn = sn
        self.name = senseid_type_config.name
        self.description = senseid_type_config.description
        self.data = []
        try:
            for data_config in senseid_type_config.data_def:
                value_raw = None
                if data_config.type == SenseidValueType.PADDING:
                    senseid_value_bytes = senseid_value_bytes[1:]
                    continue
                if data_config.type == SenseidValueType.UINT16:
                    value_raw = struct.unpack('<H', senseid_value_bytes[:2])[0]
                    senseid_value_bytes = senseid_value_bytes[2:]
                if data_config.type == SenseidValueType.INT16:
                    value_raw = struct.unpack('<h', senseid_value_bytes[:2])[0]
                    senseid_value_bytes = senseid_value_bytes[2:]
                if data_config.type == SenseidValueType.UINT16BE:
                    value_raw = struct.unpack('>H', senseid_value_bytes[:2])[0]
                    senseid_value_bytes = senseid_value_bytes[2:]
                if data_config.type == SenseidValueType.INT16BE:
                    value_raw = struct.unpack('>h', senseid_value_bytes[:2])[0]
                    senseid_value_bytes = senseid_value_bytes[2:]
                if data_config.type == SenseidValueType.FLOAT:
                    value_raw = struct.unpack('<f', senseid_value_bytes[:4])[0]
                    senseid_value_bytes = senseid_value_bytes[4:]

                value = None
                if value_raw is not None:
                    if data_config.transform == SenseidTransformType.NONE:
                        value = value_raw
                    if data_config.transform == SenseidTransformType.LINEAR:
                        value = data_config.coefficients[0] + data_config.coefficients[1] * value_raw
                    if data_config.transform == SenseidTransformType.THERMISTOR_BETA:
                        # value_raw is in adc12 value of 10k half bridge
                        r_thermistor = value_raw * 10e3 / (4095 - value_raw)
                        beta = data_config.coefficients[0]
                        r0 = data_config.coefficients[1]
                        t0 = data_config.coefficients[2] + 273.15
                        value = 1 / (1 / t0 + 1 / beta * log(r_thermistor / r0)) - 273.15

                data = SenseidData(magnitude=data_config.magnitude,
                                   unit_long=data_config.unit_long,
                                   unit_short=data_config.unit_short,
                                   value=value)
                self.data.append(data)
        except Exception as e:
            raise Exception("Error parsing senseid data")

    def parse_beacon(self, beacon: str | bytearray):
        beacon_bytes = self._get_bytearray_beacon(beacon)
        if self._is_senseid_beacon(beacon_bytes):
            self._parse_senseid_beacon(beacon_bytes)
        else:
            self.id = beacon_bytes.hex().upper()
            self.fw_version = None
            self.sn = None
            self.name = 'BLE beacon'
            self.description = 'Standard BLE beacon'
            self.data = None
        logger.debug('Parsing done -> ' + str(self))



from dataclasses import dataclass
from enum import Enum
from typing import List, Callable
from abc import ABC, abstractmethod

from dataclasses_json import dataclass_json

from ..parsers import SenseidTag


class SupportedSenseidReader(Enum):
    REDRCP = 'REDRCP'
    NURAPI = 'NURAPI'
    NURAPY = 'NURAPY'
    OCTANE = 'OCTANE'
    SPEEDWAY = 'SPEEDWAY'
    KLSBLELCR = 'KLSBLELCR'
    ACR1552 = 'ACR1552'


@dataclass_json
@dataclass
class SenseidReaderConnectionInfo:
    driver: SupportedSenseidReader
    connection_string: str


@dataclass_json
@dataclass
class SenseidReaderDetails:
    model_name: str = None
    region: str = None
    firmware_version: str = None
    antenna_count: int = None
    min_tx_power: float = None
    max_tx_power: float = None


class SenseidReader(ABC):

    @abstractmethod
    def connect(self, connection_string: str):
        pass

    @abstractmethod
    def disconnect(self):
        pass

    @abstractmethod
    def get_details(self) -> SenseidReaderDetails:
        pass

    @abstractmethod
    def get_tx_power(self) -> float:
        pass

    @abstractmethod
    def set_tx_power(self, dbm: float):
        pass

    @abstractmethod
    def get_antenna_config(self) -> List[bool]:
        pass

    @abstractmethod
    def set_antenna_config(self, antenna_config_array: List[bool]):
        pass

    @abstractmethod
    def start_inventory_async(self, notification_callback: Callable[[SenseidTag], None]):
        pass

    @abstractmethod
    def stop_inventory_async(self):
        pass


def get_supported_readers():
    return [reader.value for reader in SupportedSenseidReader]


def create_SenseidReader(reader_info: SenseidReaderConnectionInfo = None, notification_callback=None) -> SenseidReader:
    if reader_info.driver == SupportedSenseidReader.REDRCP:
        from .redrcp import SenseidReaderRedRcp
        return SenseidReaderRedRcp()
    if reader_info.driver == SupportedSenseidReader.OCTANE:
        from .octane import SenseidOctane
        return SenseidOctane()
    if reader_info.driver == SupportedSenseidReader.NURAPI:
        from .nurapi import SenseidNurapi
        return SenseidNurapi()
    if reader_info.driver == SupportedSenseidReader.NURAPY:
        from .nurapy import SenseidNurapy
        return SenseidNurapy()
    if reader_info.driver == SupportedSenseidReader.SPEEDWAY:
        from .llrp import SenseidLlrp
        return SenseidLlrp(is_impinj=True)
    if reader_info.driver == SupportedSenseidReader.KLSBLELCR:
        from .klsblelcr import SenseidKlSbleLcr
        return SenseidKlSbleLcr()
    if reader_info.driver == SupportedSenseidReader.ACR1552:
        from .acr1552 import SenseidAcr1552
        return SenseidAcr1552()



import logging
import platform
import time
from threading import Thread
from typing import Callable

import serial
import serial.tools.list_ports
from usbmonitor import USBMonitor

from .. import SenseidReaderConnectionInfo, SupportedSenseidReader

logger = logging.getLogger(__name__)


class SerialPortScanner:

    def __init__(self, notification_callback: Callable[[SenseidReaderConnectionInfo], None],
                 removal_callback: Callable[[SenseidReaderConnectionInfo], None] = None):
        self.notification_callback = notification_callback
        self.removal_callback = removal_callback
        self._scan_thread = None
        self.comports = {}  # port -> SenseidReaderConnectionInfo
        self._is_on = False

    def start(self, reset: bool = False):
        if reset:
            self.comports = {}
        self._is_on = True
        self._scan_thread = Thread(target=self._scan_job, daemon=True)
        self._scan_thread.start()

    def stop(self):
        self._is_on = False
        self._scan_thread.join()

    def _scan_job(self):

        cp_monitor = USBMonitor(filter_devices=([{'ID_VENDOR_ID': '10C4', 'ID_MODEL_ID': 'EA60'}]))

        while self._is_on:
            # Update COM ports
            com_port_list = serial.tools.list_ports.comports()
            current_ports = set()

            # Specific VIP-PID devices
            for com_port in com_port_list:
                # NUR
                if 'VID:PID=04E6:0112' in str(com_port.hwid):
                    current_ports.add(com_port.name)
                    if com_port.name not in self.comports:
                        logger.info('New NUR reader found: ' + com_port.name)
                        conn_info = SenseidReaderConnectionInfo(driver=SupportedSenseidReader.NURAPY,
                                                                connection_string=com_port.name)
                        self.comports[com_port.name] = conn_info
                        self.notification_callback(conn_info)

            # CP based COM, with device name in Serial String
            if platform.system() == 'Windows':
                cp_device_dict = cp_monitor.get_available_devices()
                for cp_device in cp_device_dict:
                    info = cp_device_dict[cp_device]
                    # SBLE-LCR
                    if info['ID_SERIAL'].startswith('KL-SBLE-LCR'):
                        port = info['ID_MODEL'].split('(')[-1].strip(')')
                        current_ports.add(port)
                        if port not in self.comports:
                            logger.info('New KL-SBLE-LCR found: ' + port
                                        + ' (SN:' + info['ID_SERIAL']
                                        + ')')
                            conn_info = SenseidReaderConnectionInfo(driver=SupportedSenseidReader.KLSBLELCR,
                                                                    connection_string=port)
                            self.comports[port] = conn_info
                            self.notification_callback(conn_info)
            else:
                for com_port in com_port_list:
                    if 'KL-SBLE-LCR' in str(com_port.serial_number):
                        current_ports.add(com_port.device)
                        if com_port.device not in self.comports:
                            logger.info('New SBLE-LCR found: ' + com_port.product
                                        + ' (SN:' + com_port.serial_number
                                        + ')')
                            conn_info = SenseidReaderConnectionInfo(driver=SupportedSenseidReader.KLSBLELCR,
                                                                    connection_string=com_port.device)
                            self.comports[com_port.device] = conn_info
                            self.notification_callback(conn_info)

            # Detect removed readers
            removed_ports = set(self.comports.keys()) - current_ports
            for port in removed_ports:
                logger.info('Reader disconnected: ' + port)
                conn_info = self.comports.pop(port)
                if self.removal_callback is not None:
                    self.removal_callback(conn_info)

            time.sleep(1)



import logging
import logging.config
import time

from src.senseid.parsers import SenseidTag
from src.senseid.readers import SupportedSenseidReader, create_SenseidReader
from src.senseid.readers.scanner import SenseidReaderScanner

logging.basicConfig(level=logging.INFO)

scanner = SenseidReaderScanner(autostart=True)
connection_info = scanner.wait_for_reader_of_type(SupportedSenseidReader.NURAPY, timeout_s=5)

if connection_info is None:
    print('No reader found')
    exit()

sid_reader = create_SenseidReader(connection_info)
sid_reader.connect(connection_info.connection_string)
#sid_reader.driver.SetLogLevel(NUR_LOG.NUR_LOG_ALL)

logging.info('Setting antenna configuration')
sid_reader.set_antenna_config(antenna_config_array=[True])

logging.info('Setting valid TX power')
sid_reader.set_tx_power(15)
logging.info('Setting too low TX power')
sid_reader.set_tx_power(sid_reader.get_details().min_tx_power - 10)
logging.info('Setting too high TX power')
sid_reader.set_tx_power(sid_reader.get_details().max_tx_power + 10)

logging.info('Setting max TX power')
sid_reader.set_tx_power(sid_reader.get_details().max_tx_power)

n_tags = 0
start = time.monotonic()
def notification_callback(tag: SenseidTag):
    global n_tags, start
    logging.info(tag)
    n_tags += 1
    print(n_tags/(time.monotonic() - start))


logging.info('Starting inventory')
sid_reader.start_inventory_async(notification_callback=notification_callback)

input()
time.sleep(1)

logging.info('Stopping inventory')
sid_reader.stop_inventory_async()

logging.info('Disconnecting from reader')
sid_reader.disconnect()



import logging
import time
from enum import Enum
from typing import List, Callable

from sllurp.llrp import LLRPReaderClient, LLRPReaderConfig, LLRP_DEFAULT_PORT, LLRPReaderState

from . import SenseidReader, SenseidReaderDetails
from ..parsers import SenseidTag
from ..parsers.rain import SenseidRainTag

logger = logging.getLogger(__name__)


class LlrpReaderManufacturer(Enum):
    IMPINJ = 25882
    MOTOROLA = 161
    SIRIT = 24831


class LlrpCommunicationStandard(Enum):
    UNSPECIFIED = 0
    US_FCC_Part_15_4W = 1
    ETSI_302_208 = 2
    ETSI_300_220_500mW = 3
    Australia_LIPD_1W = 4
    Australia_LIPD_4W = 5
    Japan_ARIB_STD_T89 = 6
    Hong_Kong_OFTA_1049_2W = 7
    Taiwan_DGT_LP0002 = 8
    Korea_MIC_Article_5_2 = 9
    _902_928MHz_4_W = 10
    ETSI_302_208_Lower_Band_2W = 11
    Brazil_Lower_Band_4W = 12
    China_Lower_Band_2W = 13
    China_Higher_Band_2W = 14
    Hong_Kong_China_4W = 15
    Israel_2W = 16
    Japan_954_4W = 17
    Japan_955_20mW = 18
    _865_868MHz_500mW = 19
    Korea_4W = 20
    Korea_200mW = 21
    Malaysia_2W = 23
    New_Zealand_Lower_Band_6W = 24
    Singapore_500mW = 25
    Singapore_2W = 26
    South_Africa_4W_FHSS = 27
    South_Africa_4W = 28
    Taiwan_1W = 29
    Taiwan_500mW = 30
    Thailand_4W = 31
    Venezuela_4W = 32
    Vietnam_500mW = 33
    Vietnam_2W = 34
    Japan_4W = 35
    Japan_500mW = 36
    Brazil_Higher_Band_4W = 37
    New_Zealand_Higher_Band_6W = 38


class ImpinjReaderMode(Enum):
    MaxThroughput = 0
    Hybrid = 1
    DenseReaderM4 = 2
    DenseReaderM8 = 3
    MaxMiller = 4
    DenseReaderM4Two = 5
    AutoSetDenseReader = 1000
    AutoSetDenseReaderDeepScan = 1002
    AutoSetStaticFast = 1003
    AutoSetStaticDRM = 1004
    AutoSetCustom = 1005


class ImpinjSearchMode(Enum):
    ReaderSelected = 0
    SingleTarget = 1
    DualTarget = 2
    TagFocus = 3
    SingleTargetReset = 5
    DualTargetBtoASelect = 6


class SenseidLlrp(SenseidReader):

    def __init__(self, is_impinj: bool):
        self.driver: LLRPReaderClient | None = None
        self.notification_callback = None
        self.details = SenseidReaderDetails()
        self._tx_power: float | None = None
        self._antennas: List[bool] = [True]
        self._is_impinj: bool = is_impinj

    def connect(self, connection_string: str):
        try:
            self.connection_string = connection_string
            config = LLRPReaderConfig()
            config.start_inventory = False
            config.reset_on_connect = True
            if self._is_impinj:
                config.impinj_extended_configuration = True
            self.driver = LLRPReaderClient(self.connection_string, LLRP_DEFAULT_PORT, config)

            capabilities_received = False

            def on_get_capabilities(reader: LLRPReaderClient, state):
                nonlocal capabilities_received
                if not capabilities_received:
                    logger.debug('TODO: parse capabilities with non impinj readers')

                    if self._is_impinj:
                        self.details.model_name = self.driver.llrp.capabilities['ImpinjDetailedVersion'][
                        'ModelName'].decode('utf-8')
                        self.details.firmware_version = self.driver.llrp.capabilities['ImpinjDetailedVersion'][
                        'FirmwareVersion'].decode('utf-8')
                    else:
                        self.details.model_name = self.driver.llrp.capabilities['GeneralDeviceCapabilities'][
                        'ModelName']
                        self.details.firmware_version = self.driver.llrp.capabilities['GeneralDeviceCapabilities'][
                        'ReaderFirmwareVersion']
                    self.details.antenna_count = self.driver.llrp.capabilities['GeneralDeviceCapabilities'][
                        'MaxNumberOfAntennaSupported']
                    self.details.min_tx_power = self.driver.llrp.tx_power_table[1]
                    self.details.max_tx_power = self.driver.llrp.tx_power_table[-1]
                    self.details.region = LlrpCommunicationStandard(
                        self.driver.llrp.capabilities['RegulatoryCapabilities']['CommunicationsStandard']).name
                    logger.debug(reader.llrp.capabilities)
                    capabilities_received = True

            self.driver.add_state_callback(LLRPReaderState.STATE_SENT_GET_CONFIG, on_get_capabilities)
            self.driver.add_tag_report_callback(self._llrp_notification_callback)
            self.driver.connect()

            while not capabilities_received:
                time.sleep(0.1)
            self.driver.disconnect(None)
            self._tx_power = self.details.max_tx_power

            return True
        except Exception as e:
            logger.error(e)
            return False

    def _llrp_notification_callback(self, reader: LLRPReaderClient, tag_reports):
        logger.debug(tag_reports)
        if self.notification_callback is not None:
            for tag in tag_reports:
                #for n in range(tag['TagSeenCount']):
                self.notification_callback(SenseidRainTag(epc=tag['EPC'].decode('utf-8')))

    def disconnect(self):
        self.driver.disconnect()
        self.driver.join()

    def get_details(self) -> SenseidReaderDetails:
        return self.details

    def get_tx_power(self) -> float:
        # Only supporting same power on all antennas
        return self._tx_power

    def set_tx_power(self, dbm: float):
        # Only supporting same power on all antennas
        if self.details is None:
            self.get_details()
        if dbm > self.details.max_tx_power:
            dbm = self.details.max_tx_power
            logger.warning('Power set to max power: ' + str(dbm))
        if dbm < self.details.min_tx_power:
            dbm = self.details.min_tx_power
            logger.warning('Power set to min power: ' + str(dbm))
        self._tx_power = dbm

    def get_antenna_config(self) -> List[bool]:
        return self._antennas

    def set_antenna_config(self, antenna_config_array: List[bool]):
        if not (True in antenna_config_array):
            antenna_config_array[0] = True
            logger.warning('At least one antenna needs to be active. Enabling antenna 1.')
        self._antennas = antenna_config_array

    def start_inventory_async(self, notification_callback: Callable[[SenseidTag], None]):
        self.notification_callback = notification_callback

        config = LLRPReaderConfig()
        config.start_inventory = True
        config.reset_on_connect = True
        if self._is_impinj:
            config.impinj_extended_configuration = True
        config.duration = 0.2
        config.tag_content_selector = {
            'EnableROSpecID': False,
            'EnableSpecIndex': False,
            'EnableInventoryParameterSpecID': False,
            'EnableAntennaID': False,
            'EnableChannelIndex': False,
            'EnablePeakRSSI': False,
            'EnableFirstSeenTimestamp': False,
            'EnableLastSeenTimestamp': False,
            'EnableTagSeenCount': True,
            'EnableAccessSpecID': False,
            'C1G2EPCMemorySelector': {
                'EnableCRC': False,
                'EnablePCBits': False,
            }
        }
        if self._is_impinj:
            config.impinj_search_mode = ImpinjReaderMode.DenseReaderM4.value
        config.mode_identifier = ImpinjSearchMode.DualTarget.value
        config.antennas = []
        config.tx_power = {}
        config.tx_power_dbm = {}
        for idx, enable in enumerate(self._antennas):
            if enable:
                config.antennas.append(idx + 1)
                config.tx_power[idx + 1] = 0
                config.tx_power_dbm[idx + 1] = self._tx_power

        self.driver = LLRPReaderClient(self.connection_string, LLRP_DEFAULT_PORT, config)
        self.driver.add_tag_report_callback(self._llrp_notification_callback)
        self.driver.connect()

    def stop_inventory_async(self):
        config = LLRPReaderConfig()
        config.start_inventory = False
        config.reset_on_connect = True
        self.driver.update_config(config)
        self.driver.disconnect(None)



import logging

from driver_snfc_py_acr1552.acr1552 import Acr1552

from . import SenseidReaderDetails
from ..parsers.nfc import SenseidNfcTag, convert_to_uint, Endianness

logger = logging.getLogger(__name__)


class SenseidAcr1552:

    NTAG5_NDEF_BASE_BLOCK = 0
    NTAG5_NDEF_HEADER_NBLOCKS = 2

    NTAG5_IDX_BLOCK = 50
    NTAG5_IDX_NBLOCKS = 1

    NTAG5_DATA_BASE_BLOCK = 0
    NTAG5_DATA_NBLOCKS = 50

    def __init__(self):
        self.driver = Acr1552()
        self.details = None

    def connect(self, connection_string: str = None):
        return self.driver.connect(connection_string=connection_string)

    def disconnect(self):
        self.driver.disconnect()

    def get_details(self) -> SenseidReaderDetails:
        if self.details is None:
            self.details = SenseidReaderDetails(
                model_name='ACR1552',
                region='EU',
                firmware_version='1.0.1',
                antenna_count=1,
                min_tx_power=0,
                max_tx_power=0
            )
        return self.details

    # -- Field control --

    def turn_on_field(self):
        return self.driver.set_power(True)

    def turn_off_field(self):
        return self.driver.set_power(False)

    # -- Mode control --

    def set_ndef_mode(self):
        return self.driver.change_fw_mode(False)

    def set_bulk_mode(self):
        return self.driver.change_fw_mode(True)

    # -- Tag operations --

    def get_uid(self):
        return self.driver.get_uid()

    def read_ndef(self, uid=None) -> SenseidNfcTag:
        # First read: get CC + NDEF TLV header to determine message length
        header = self.driver.read_data(self.NTAG5_NDEF_BASE_BLOCK, self.NTAG5_NDEF_HEADER_NBLOCKS)
        if header is None or len(header) < 6:
            return None
        # Verify CC (E1 40) and NDEF TLV type (03)
        if header[0] != 0xE1 or header[1] != 0x40 or header[4] != 0x03:
            return None
        # NDEF length at byte 5: total bytes = CC(4) + TLV type(1) + TLV length(1) + payload
        ndef_length = header[5]
        total_bytes = 4 + 1 + 1 + ndef_length
        total_blocks = (total_bytes + 3) // 4  # Round up to full blocks
        # Second read: get the full NDEF data
        data = self.driver.read_data(self.NTAG5_NDEF_BASE_BLOCK, total_blocks)
        if data is None:
            return None
        uid_str = bytearray(uid).hex().upper() if uid else None
        return SenseidNfcTag(bytearray(data), uid=uid_str)

    def read_tag_index(self):
        data = self.driver.read_data(self.NTAG5_IDX_BLOCK, self.NTAG5_IDX_NBLOCKS)
        idx_list = convert_to_uint(data, 4, Endianness.LITTLE.value)
        if idx_list:
            return idx_list[0]
        return None

    def read_bulk_data(self):
        data = self.driver.read_data(self.NTAG5_DATA_BASE_BLOCK, self.NTAG5_DATA_NBLOCKS)
        if data is None:
            return None
        return convert_to_uint(data, 2, Endianness.LITTLE.value)

    # -- Raw access --

    def read_data(self, start_block, n_blocks):
        return self.driver.read_data(start_block, n_blocks)

    def write_data(self, start_block, wdata):
        return self.driver.write_data(start_block, wdata)


import logging
from typing import Callable

from zeroconf import IPVersion, Zeroconf, ServiceBrowser, ServiceStateChange
from .. import SenseidReaderConnectionInfo, SupportedSenseidReader

logger = logging.getLogger(__name__)


class MulticastDnsServiceDiscoveryScanner:

    def __init__(self, notification_callback: Callable[[SenseidReaderConnectionInfo], None],
                 removal_callback: Callable[[SenseidReaderConnectionInfo], None] = None,
                 autostart: bool = False):
        self.notification_callback = notification_callback
        self.removal_callback = removal_callback
        self.service_browser: ServiceBrowser | None = None
        self.zeroconf_instance = Zeroconf(ip_version=IPVersion.V4Only)
        self.ips = {}  # ip_str -> list of SenseidReaderConnectionInfo

        if autostart:
            self.start()

    def start(self, reset: bool = False):
        if reset:
            self.ips = {}

        def on_service_state_change(zeroconf: Zeroconf, service_type: str, name: str, state_change: ServiceStateChange):
            if 'SpeedwayR' in name:
                info = zeroconf.get_service_info(service_type, name)
                if info is not None:
                    ipv4_addresses = info.addresses_by_version(IPVersion.V4Only)
                    if len(ipv4_addresses) > 0:
                        ipv4 = ipv4_addresses[0]
                        ip_str = str(ipv4[0]) + '.' + str(ipv4[1]) + '.' + str(ipv4[2]) + '.' + str(ipv4[3])
                        if state_change is ServiceStateChange.Added:
                            if ip_str not in self.ips:
                                logger.info('New Speedway readers found: ' + ip_str)
                                conn_infos = [
                                    SenseidReaderConnectionInfo(driver=SupportedSenseidReader.OCTANE,
                                                                connection_string=ip_str),
                                    SenseidReaderConnectionInfo(driver=SupportedSenseidReader.SPEEDWAY,
                                                                connection_string=ip_str),
                                ]
                                self.ips[ip_str] = conn_infos
                                for conn_info in conn_infos:
                                    self.notification_callback(conn_info)
                        elif state_change is ServiceStateChange.Removed:
                            if ip_str in self.ips:
                                logger.info('Speedway reader disconnected: ' + ip_str)
                                conn_infos = self.ips.pop(ip_str)
                                if self.removal_callback is not None:
                                    for conn_info in conn_infos:
                                        self.removal_callback(conn_info)

            if 'ThingMagic Mercury' in name:
                info = zeroconf.get_service_info(service_type, name)
                if info is not None:
                    ipv4_addresses = info.addresses_by_version(IPVersion.V4Only)
                    if len(ipv4_addresses) > 0:
                        ipv4 = ipv4_addresses[0]
                        ip_str = str(ipv4[0]) + '.' + str(ipv4[1]) + '.' + str(ipv4[2]) + '.' + str(ipv4[3])
                        if state_change is ServiceStateChange.Added:
                            if ip_str not in self.ips:
                                logger.info('New Mercury readers found: ' + ip_str)
                                conn_infos = [
                                    SenseidReaderConnectionInfo(driver=SupportedSenseidReader.SPEEDWAY,
                                                                connection_string=ip_str),
                                ]
                                self.ips[ip_str] = conn_infos
                                for conn_info in conn_infos:
                                    self.notification_callback(conn_info)
                        elif state_change is ServiceStateChange.Removed:
                            if ip_str in self.ips:
                                logger.info('Mercury reader disconnected: ' + ip_str)
                                conn_infos = self.ips.pop(ip_str)
                                if self.removal_callback is not None:
                                    for conn_info in conn_infos:
                                        self.removal_callback(conn_info)

        services = [
            "_http._tcp.local.",
        ]
        self.service_browser = ServiceBrowser(self.zeroconf_instance, services, handlers=[on_service_state_change])

    def stop(self):
        self.service_browser.cancel()
        self.service_browser.join()



"""
Script de test para ACR1552 (funciones de alto nivel).
Usa el scanner del SDK para detectar el lector y las funciones bridge de SenseidAcr1552.
"""

import logging
from time import sleep

from senseid.readers import SupportedSenseidReader, create_SenseidReader
from senseid.readers.scanner import SenseidReaderScanner
from senseid.parsers.nfc import SenseidNfcTag

logging.basicConfig(level=logging.DEBUG, format="%(asctime)s [%(levelname)s] %(message)s")
logger = logging.getLogger(__name__)


# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------

def bytes_to_hex(data):
    if data is None:
        return "None"
    return " ".join(f"{b:02X}" for b in data)


# ---------------------------------------------------------------------------
# Test blocks
# ---------------------------------------------------------------------------

def test_turn_on_field(reader):
    print("\n--- turn_on_field() ---")
    ok = reader.turn_on_field()
    print(f"  Result: {'OK' if ok else 'FAIL'}")
    return ok


def test_turn_off_field(reader):
    print("\n--- turn_off_field() ---")
    ok = reader.turn_off_field()
    print(f"  Result: {'OK' if ok else 'FAIL'}")
    return ok


def test_get_uid(reader):
    print("\n--- get_uid() ---")
    uid = reader.get_uid()
    if uid:
        print(f"  UID: {bytes_to_hex(uid)}")
    else:
        print("  No tag detected (UID = None)")
    return uid


def test_read_data(reader, start_block, n_blocks):
    print(f"\n--- read_data(start={start_block}, blocks={n_blocks}) ---")
    data = reader.read_data(start_block, n_blocks)
    if data:
        print(f"  Raw ({len(data)} bytes): {bytes_to_hex(data)}")
    else:
        print("  No data (None)")
    return data


def test_read_ndef(reader):
    print("\n--- read_ndef() ---")
    uid = reader.get_uid()
    nfc_tag = reader.read_ndef(uid=uid)
    if nfc_tag is not None:
        print(f"  ID:          {nfc_tag.id}")
        print(f"  Name:        {nfc_tag.name}")
        print(f"  Description: {nfc_tag.description}")
        if nfc_tag.data:
            for d in nfc_tag.data:
                print(f"  {d.magnitude}: {d.value} {d.unit_short}")
        else:
            print("  No sensor data parsed")
    else:
        print("  Could not read NDEF")
    return nfc_tag


def test_read_tag_index(reader):
    print("\n--- read_tag_index() ---")
    idx = reader.read_tag_index()
    if idx is not None:
        print(f"  Index (uint32): {idx}")
    else:
        print("  Could not read index")
    return idx


def test_read_bulk_data(reader):
    print("\n--- read_bulk_data() ---")
    values = reader.read_bulk_data()
    if values:
        print(f"  Total uint16 values: {len(values)}")
        print(f"  First 20: {values[:20]}")
    else:
        print("  Could not read bulk data")
    return values


def test_set_ndef_mode(reader):
    print("\n--- set_ndef_mode() ---")
    ok = reader.set_ndef_mode()
    print(f"  Result: {'OK' if ok else 'FAIL'}")
    return ok


def test_set_bulk_mode(reader):
    print("\n--- set_bulk_mode() ---")
    ok = reader.set_bulk_mode()
    print(f"  Result: {'OK' if ok else 'FAIL'}")
    return ok


def test_write_read_back(reader, block=48):
    test_data = [0xDE, 0xAD, 0xBE, 0xEF]
    print(f"\n--- write_data(block={block}, data={bytes_to_hex(test_data)}) ---")
    ok = reader.write_data(block, test_data)
    print(f"  Write: {'OK' if ok else 'FAIL'}")
    if ok:
        readback = reader.read_data(block, 1)
        print(f"  Read back: {bytes_to_hex(readback)}")
        if readback and list(readback[:4]) == test_data:
            print("  Verification OK")
        else:
            print("  Verification FAILED - data mismatch")
    return ok


# ---------------------------------------------------------------------------
# Full test sequence
# ---------------------------------------------------------------------------

def run_full_test(reader):
    print("\n" + "=" * 60)
    print("  RUNNING FULL TEST")
    print("=" * 60)

    # 1. RF ON
    test_turn_on_field(reader)
    sleep(0.3)

    # 2. UID
    uid = test_get_uid(reader)
    if uid is None:
        print("\n  No tag present. Place a tag and try again.")
        return

    # 3. NDEF mode - read message
    test_set_ndef_mode(reader)
    sleep(0.3)
    test_turn_on_field(reader)
    sleep(0.3)
    test_read_ndef(reader)

    # 4. BULK mode - read index and data
    test_set_bulk_mode(reader)
    sleep(0.3)
    test_turn_on_field(reader)
    sleep(0.3)
    test_read_tag_index(reader)
    test_read_bulk_data(reader)

    # 5. Write/read test
    test_write_read_back(reader, block=48)

    # 6. Back to NDEF
    test_set_ndef_mode(reader)

    print("\n" + "=" * 60)
    print("  FULL TEST COMPLETE")
    print("=" * 60)


# ---------------------------------------------------------------------------
# Interactive menu
# ---------------------------------------------------------------------------

def print_menu():
    print("\n" + "=" * 60)
    print("  TEST ACR1552 (High-Level) - Menu")
    print("=" * 60)
    print("  1. Turn ON RF field")
    print("  2. Turn OFF RF field")
    print("  3. Read tag UID")
    print("  4. Read arbitrary block(s)")
    print("  5. Read NDEF message")
    print("  6. Read BULK index")
    print("  7. Read BULK data")
    print("  8. Set BULK mode")
    print("  9. Set NDEF mode")
    print(" 10. Write and verify test block")
    print(" 11. Run full test sequence")
    print("  0. Exit")
    print("-" * 60)


def run_interactive(reader):
    while True:
        print_menu()
        try:
            choice = input("Select option: ").strip()
        except (EOFError, KeyboardInterrupt):
            break

        if choice == "1":
            test_turn_on_field(reader)
        elif choice == "2":
            test_turn_off_field(reader)
        elif choice == "3":
            test_get_uid(reader)
        elif choice == "4":
            try:
                sb = int(input("  Start block: "))
                nb = int(input("  Number of blocks: "))
                test_read_data(reader, sb, nb)
            except ValueError:
                print("  Invalid input")
        elif choice == "5":
            test_read_ndef(reader)
        elif choice == "6":
            test_read_tag_index(reader)
        elif choice == "7":
            test_read_bulk_data(reader)
        elif choice == "8":
            test_set_bulk_mode(reader)
        elif choice == "9":
            test_set_ndef_mode(reader)
        elif choice == "10":
            test_write_read_back(reader)
        elif choice == "11":
            run_full_test(reader)
        elif choice == "0":
            break
        else:
            print("  Invalid option")


# ---------------------------------------------------------------------------
# Entry point
# ---------------------------------------------------------------------------

if __name__ == "__main__":
    print("Scanning for ACR1552 reader...")
    scanner = SenseidReaderScanner(autostart=True)
    connection_info = scanner.wait_for_reader_of_type(SupportedSenseidReader.ACR1552, timeout_s=10)

    if connection_info is None:
        print("No ACR1552 reader found. Is it connected?")
        exit(1)

    print(f"Found reader: {connection_info.connection_string}")
    sid_reader = create_SenseidReader(connection_info)
    sid_reader.connect(connection_info.connection_string)
    scanner.stop()

    try:
        run_interactive(sid_reader)
    except Exception as e:
        logger.exception(f"Unexpected error: {e}")
    finally:
        print("\nDisconnecting...")
        try:
            sid_reader.turn_on_field()
            sid_reader.disconnect()
        except Exception:
            pass
        print("Done.")



import logging
import time
from threading import Thread
from typing import Callable

from smartcard.System import readers

from .. import SenseidReaderConnectionInfo, SupportedSenseidReader

logger = logging.getLogger(__name__)


class PcscScanner:

    def __init__(self, notification_callback: Callable[[SenseidReaderConnectionInfo], None],
                 removal_callback: Callable[[SenseidReaderConnectionInfo], None] = None):
        self.notification_callback = notification_callback
        self.removal_callback = removal_callback
        self._scan_thread = None
        self.found_readers = {}  # reader_name -> SenseidReaderConnectionInfo
        self._is_on = False

    def start(self, reset: bool = False):
        if reset:
            self.found_readers = {}
        self._is_on = True
        self._scan_thread = Thread(target=self._scan_job, daemon=True)
        self._scan_thread.start()

    def stop(self):
        self._is_on = False
        if self._scan_thread is not None:
            self._scan_thread.join()

    def _scan_job(self):
        while self._is_on:
            try:
                pcsc_readers = readers()
                current_readers = set()
                for reader in pcsc_readers:
                    reader_name = str(reader)
                    if 'ACR1552' in reader_name and 'PICC' in reader_name:
                        current_readers.add(reader_name)
                        if reader_name not in self.found_readers:
                            logger.info('New ACR1552 reader found: ' + reader_name)
                            conn_info = SenseidReaderConnectionInfo(
                                driver=SupportedSenseidReader.ACR1552,
                                connection_string=reader_name
                            )
                            self.found_readers[reader_name] = conn_info
                            self.notification_callback(conn_info)
                # Detect removed readers
                removed_readers = set(self.found_readers.keys()) - current_readers
                for reader_name in removed_readers:
                    logger.info('ACR1552 reader disconnected: ' + reader_name)
                    conn_info = self.found_readers.pop(reader_name)
                    if self.removal_callback is not None:
                        self.removal_callback(conn_info)
            except Exception as e:
                logger.debug(f"PC/SC scan error: {e}")
            time.sleep(1)


import logging
from typing import List, Callable

from nurapy import NurAPY, NurTagDataMeta, InventoryStreamNotification, ModuleSetupFlags, ModuleSetup, NurDeviceCaps
from nurapy.protocol.command.module_setup import ModuleSetupLinkFreq, ModuleSetupRxDec

from . import SenseidReader, SenseidReaderDetails
from ..parsers import SenseidTag
from ..parsers.rain import SenseidRainTag

logger = logging.getLogger(__name__)


class SenseidNurapy(SenseidReader):

    def __init__(self):
        self.driver = NurAPY()
        self.notification_callback = None
        self.device_caps: NurDeviceCaps | None = None
        self.details = None

    def connect(self, connection_string: str):
        self.driver.connect(connection_string=connection_string)
        self.driver.set_notification_callback(self._nur_notification_callback)
        self.get_details()

        # Set Senseid compatible mode
        module_setup = ModuleSetup()
        module_setup.link_freq = ModuleSetupLinkFreq.BLF_256
        module_setup.rx_decoding = ModuleSetupRxDec.MILLER_4
        self.driver.set_module_setup(setup_flags=[ModuleSetupFlags.LINKFREQ,
                                                  ModuleSetupFlags.RXDEC],
                                     module_setup=module_setup)

        # Set MAX TX Power
        self.set_tx_power(self.details.max_tx_power)

        # Enable first antenna
        antenna_config = [False] * self.details.antenna_count
        antenna_config[0] = True
        self.set_antenna_config(antenna_config_array=antenna_config)
        return True

    def _nur_notification_callback(self, inventory_stream_notification: InventoryStreamNotification,
                                   tags: List[NurTagDataMeta]):
        if inventory_stream_notification.stopped:
            logging.info('Restarting inventory stream')
            self.driver.start_inventory_stream()
        for tag in tags:
            self.notification_callback(SenseidRainTag(epc=tag.epc))
        self.driver.clear_notified_tags()

    def disconnect(self):
        self.driver.disconnect()

    def get_details(self) -> SenseidReaderDetails:
        if self.details is None:
            reader_info = self.driver.get_reader_info()
            self.device_caps = self.driver.get_device_capabilities()

            module_setup = self.driver.get_module_setup(setup_flags=[ModuleSetupFlags.REGION])

            self.details = SenseidReaderDetails(
                model_name=reader_info.name,
                region=module_setup.region_id.name,
                firmware_version=reader_info.sw_version,
                antenna_count=reader_info.num_antennas,
                min_tx_power=self.device_caps.maxTxdBm - (self.device_caps.txSteps - 1) * self.device_caps.txAttnStep,
                max_tx_power=self.device_caps.maxTxdBm
            )
            logger.debug(self.details)
        return self.details

    def get_tx_power(self) -> float:
        # Only supporting same power on all antennas
        module_setup = self.driver.get_module_setup(setup_flags=[ModuleSetupFlags.TXLEVEL])
        current_tx_dbm = self.device_caps.maxTxdBm - module_setup.tx_level / self.device_caps.txAttnStep
        logger.debug('get_tx_power: ' + str(current_tx_dbm))
        return current_tx_dbm

    def set_tx_power(self, dbm: float):
        logger.debug('set_tx_power: ' + str(dbm))
        # Only supporting same power on all antennas
        if self.details is None:
            self.get_details()
        if dbm > self.details.max_tx_power:
            dbm = self.details.max_tx_power
            logger.warning('Power set to max power: ' + str(dbm))
        if dbm < self.details.min_tx_power:
            dbm = self.details.min_tx_power
            logger.warning('Power set to min power: ' + str(dbm))

        module_setup = ModuleSetup()
        module_setup.tx_level = int((self.device_caps.maxTxdBm - dbm) * self.device_caps.txAttnStep)
        self.driver.set_module_setup(setup_flags=[ModuleSetupFlags.TXLEVEL],
                                     module_setup=module_setup)

    def get_antenna_config(self) -> List[bool]:
        module_setup = self.driver.get_module_setup(setup_flags=[ModuleSetupFlags.ANTMASK])
        antenna_mask = module_setup.antenna_mask
        antenna_config_array = []
        for i in range(self.details.antenna_count):
            antenna_bit = (antenna_mask >> i) & 0b1
            antenna_config_array.append(bool(antenna_bit))
        logger.debug('get_antenna_config: ' + str(antenna_config_array))
        return antenna_config_array

    def set_antenna_config(self, antenna_config_array: List[bool]):
        logger.debug('set_antenna_config: ' + str(antenna_config_array))
        if not (True in antenna_config_array):
            antenna_config_array[0] = True
            logger.warning('At least one antenna needs to be active. Enabling antenna 1.')
        antenna_mask = 0
        for idx, antenna_config in enumerate(antenna_config_array):
            antenna_mask |= antenna_config << idx
        module_setup = ModuleSetup()
        module_setup.antenna_mask = antenna_mask
        module_setup.selected_antenna = 255  # Automatic selection
        self.driver.set_module_setup(setup_flags=[ModuleSetupFlags.ANTMASK,
                                                  ModuleSetupFlags.SELECTEDANT],
                                     module_setup=module_setup)

    def start_inventory_async(self, notification_callback: Callable[[SenseidTag], None]):
        self.notification_callback = notification_callback
        return self.driver.start_inventory_stream()

    def stop_inventory_async(self):
        return self.driver.stop_inventory_stream()



import datetime
from dataclasses import dataclass
from enum import Enum
from importlib.resources import files
from typing import List, Dict

import yaml
from dataclasses_json import dataclass_json


class SenseidValueType(Enum):
    UINT16 = 'uint16'
    INT16 = 'int16'
    FLOAT = 'float'


class SenseidTransformType(Enum):
    NONE = 'none'
    LINEAR = 'linear'
    THERMISTOR_BETA = 'thermistor-beta'


@dataclass_json
@dataclass
class SenseidRainDataDef:
    magnitude: str
    unit_long: str
    unit_short: str
    type: SenseidValueType
    transform: SenseidTransformType
    coefficients: List[float]


@dataclass_json
@dataclass
class SenseidRainTypeDef:
    name: str
    description: str
    data_def: List[SenseidRainDataDef]
    fw_versions: List[int]


@dataclass_json
@dataclass
class SenseidRainDef:
    version: int
    date: datetime.date
    pen_header: bytearray
    types: Dict[int, SenseidRainTypeDef]


# Detect Source or Package mode
top_package = __name__.split('.')[0]
if top_package == 'src':
    senseid_package = files('src.senseid')
else:
    senseid_package = files('senseid')
_senseid_yaml = senseid_package.joinpath('definitions').joinpath('senseid_rain.yaml').read_text()
_senseid_dict = yaml.safe_load(_senseid_yaml)
SENSEID_RAIN_DEF: SenseidRainDef = SenseidRainDef.from_dict(_senseid_dict)



import logging
import time

from src.senseid.readers import SenseidReaderConnectionInfo, SupportedSenseidReader
from src.senseid.readers.scanner import SenseidReaderScanner

logging.basicConfig(level=logging.DEBUG)


def scanner_notification_callback(new_reader: SenseidReaderConnectionInfo):
    logging.info(new_reader)


# Scan readers with notifications
scanner = SenseidReaderScanner(notification_callback=scanner_notification_callback, autostart=True)

time.sleep(1)

# Get all readers found by the scanner
reader_list = scanner.get_readers()
logging.info(reader_list)

# Or wait until desired reader is found
nur_reader = scanner.wait_for_reader_of_type(reader_type=SupportedSenseidReader.NURAPI, timeout_s=1)
logging.info(nur_reader)
scanner.stop()
